class OVRLibException(Exception):
    pass


class InvalidAccessKeyError(OVRLibException):
    pass


class ReadOnlyAccessKeyError(OVRLibException):
    pass


class InvalidRegistrationError(OVRLibException):
    pass


class InvalidDLError(OVRLibException):
    pass


class InvalidSignatureError(OVRLibException):
    pass



import datetime
import re
from dataclasses import dataclass
from typing import Optional

import requests

QUERY_ENDPOINT = "https://www.mvp.sos.ga.gov/MVP/voterDetails.do"

COUNTIES = {
    "APPLING": "001",
    "ATKINSON": "002",
    "BACON": "003",
    "BAKER": "004",
    "BALDWIN": "005",
    "BANKS": "006",
    "BARROW": "007",
    "BARTOW": "008",
    "BEN HILL": "009",
    "BERRIEN": "010",
    "BIBB": "011",
    "BLECKLEY": "012",
    "BRANTLEY": "013",
    "BROOKS": "014",
    "BRYAN": "015",
    "BULLOCH": "016",
    "BURKE": "017",
    "BUTTS": "018",
    "CALHOUN": "019",
    "CAMDEN": "020",
    "CANDLER": "021",
    "CARROLL": "022",
    "CATOOSA": "023",
    "CHARLTON": "024",
    "CHATHAM": "025",
    "CHATTAHOOCHEE": "026",
    "CHATTOOGA": "027",
    "CHEROKEE": "028",
    "CLARKE": "029",
    "CLAY": "030",
    "CLAYTON": "031",
    "CLINCH": "032",
    "COBB": "033",
    "COFFEE": "034",
    "COLQUITT": "035",
    "COLUMBIA": "036",
    "COOK": "037",
    "COWETA": "038",
    "CRAWFORD": "039",
    "CRISP": "040",
    "DADE": "041",
    "DAWSON": "042",
    "DECATUR": "043",
    "DEKALB": "044",
    "DODGE": "045",
    "DOOLY": "046",
    "DOUGHERTY": "047",
    "DOUGLAS": "048",
    "EARLY": "049",
    "ECHOLS": "050",
    "EFFINGHAM": "051",
    "ELBERT": "052",
    "EMANUEL": "053",
    "EVANS": "054",
    "FANNIN": "055",
    "FAYETTE": "056",
    "FLOYD": "057",
    "FORSYTH": "058",
    "FRANKLIN": "059",
    "FULTON": "060",
    "GILMER": "061",
    "GLASCOCK": "062",
    "GLYNN": "063",
    "GORDON": "064",
    "GRADY": "065",
    "GREENE": "066",
    "GWINNETT": "067",
    "HABERSHAM": "068",
    "HALL": "069",
    "HANCOCK": "070",
    "HARALSON": "071",
    "HARRIS": "072",
    "HART": "073",
    "HEARD": "074",
    "HENRY": "075",
    "HOUSTON": "076",
    "IRWIN": "077",
    "JACKSON": "078",
    "JASPER": "079",
    "JEFF DAVIS": "080",
    "JEFFERSON": "081",
    "JENKINS": "082",
    "JOHNSON": "083",
    "JONES": "084",
    "LAMAR": "085",
    "LANIER": "086",
    "LAURENS": "087",
    "LEE": "088",
    "LIBERTY": "089",
    "LINCOLN": "090",
    "LONG": "091",
    "LOWNDES": "092",
    "LUMPKIN": "093",
    "MACON": "094",
    "MADISON": "095",
    "MARION": "096",
    "MCDUFFIE": "097",
    "MCINTOSH": "098",
    "MERIWETHER": "099",
    "MILLER": "100",
    "MITCHELL": "101",
    "MONROE": "102",
    "MONTGOMERY": "103",
    "MORGAN": "104",
    "MURRAY": "105",
    "MUSCOGEE": "106",
    "NEWTON": "107",
    "OCONEE": "108",
    "OGLETHORPE": "109",
    "PAULDING": "110",
    "PEACH": "111",
    "PICKENS": "112",
    "PIERCE": "113",
    "PIKE": "114",
    "POLK": "115",
    "PULASKI": "116",
    "PUTNAM": "117",
    "QUITMAN": "118",
    "RABUN": "119",
    "RANDOLPH": "120",
    "RICHMOND": "121",
    "ROCKDALE": "122",
    "SCHLEY": "123",
    "SCREVEN": "124",
    "SEMINOLE": "125",
    "SPALDING": "126",
    "STEPHENS": "127",
    "STEWART": "128",
    "SUMTER": "129",
    "TALBOT": "130",
    "TALIAFERRO": "131",
    "TATTNALL": "132",
    "TAYLOR": "133",
    "TELFAIR": "134",
    "TERRELL": "135",
    "THOMAS": "136",
    "TIFT": "137",
    "TOOMBS": "138",
    "TOWNS": "139",
    "TREUTLEN": "140",
    "TROUP": "141",
    "TURNER": "142",
    "TWIGGS": "143",
    "UNION": "144",
    "UPSON": "145",
    "WALKER": "146",
    "WALTON": "147",
    "WARE": "148",
    "WARREN": "149",
    "WASHINGTON": "150",
    "WAYNE": "151",
    "WEBSTER": "152",
    "WHEELER": "153",
    "WHITE": "154",
    "WHITFIELD": "155",
    "WILCOX": "156",
    "WILKES": "157",
    "WILKINSON": "158",
    "WORTH": "159",
}


class GAInvalidCounty(Exception):
    pass


@dataclass
class GAVoterRegistration:
    full_name: str
    address: str
    city: str
    state: str
    zipcode: str

    date_of_birth: datetime.date

    status: str

    registration_date: datetime.date

    voter_reg_number: str  # NOTE: should this be an integer?

    @property
    def active(self):
        return self.status == "Active"

    @classmethod
    def from_page_source(cls, html, dob):
        html = html.replace("\n", " ")

        def get_span_value(spanid):
            m = re.search(f"<span id=['\"]{spanid}['\"]>([^<]*)</span>", html)
            if m:
                return m[1].strip()

        try:
            m = re.search(
                f'input type="hidden" name="idVoter" id="idVoter" value="(\d+)"/>', html
            )
            if m:
                voter_reg_number = m[1]
            else:
                return None
            reg_date = get_span_value("regDtSpan").split(": ")[1]
            address1 = get_span_value("resAddress1")
            address2 = get_span_value("resAddress2") or ""

            return GAVoterRegistration(
                full_name=get_span_value("fullNameSpan"),
                address=f"{address1} {address2}".strip(),
                city=get_span_value("resAddress3"),
                state=get_span_value("resAddress4").replace(",", "").strip(),
                zipcode=get_span_value("resAddress5").replace(",", "").strip(),
                date_of_birth=dob,
                status=get_span_value("statuscontent"),
                registration_date=datetime.datetime.strptime(
                    reg_date, "%m/%d/%Y"
                ).date(),
                voter_reg_number=voter_reg_number,
            )
        except Exception:
            return None


def lookup_voter(
    first_name: str, last_name: str, date_of_birth: datetime.date, county: str, **kwargs
) -> Optional[GAVoterRegistration]:
    county_id = COUNTIES.get(county.upper())
    if not county_id:
        raise GAInvalidCounty(f"{county} is not a recognized county")
    response = requests.post(
        QUERY_ENDPOINT,
        headers={"content-type": "application/x-www-form-urlencoded",},
        data={
            "firstName": first_name,
            "lastName": last_name,
            "dob": date_of_birth.strftime("%m/%d/%Y"),
            "county": county_id,
        },
        **kwargs,
    )
    if response.status_code != 200:
        return None

    return GAVoterRegistration.from_page_source(
        response.content.decode("utf-8"), date_of_birth
    )



import datetime
from dataclasses import dataclass
from typing import Optional

import requests

SEARCH_ENDPOINT = (
    "https://myvote.wi.gov/DesktopModules/GabMyVoteModules/api/voter/search"
)
POLLING_PLACE_ENDPOINT = "https://myvote.wi.gov/DesktopModules/GabMyVoteModules/api/address/pollingplace/{district_combo_id}"
BALLOT_ENDPOINT = "https://myvote.wi.gov/DesktopModules/GabMyVoteModules/api/absentee/progressbarinfo/{voter_id}?electionid={election_id}"


@dataclass
class WIAbsenteeBallotStatus:
    request_submitted: Optional[datetime.datetime]
    request_approved: Optional[datetime.datetime]
    created: Optional[datetime.datetime]
    sent: Optional[datetime.datetime]
    expected_delivery: Optional[datetime.datetime]
    received: Optional[datetime.datetime]
    returned: Optional[datetime.datetime]

    ballot_rejected: Optional[bool]
    request_denied: Optional[bool]
    ballot_received_issue: Optional[str]
    foreign_address: Optional[bool]

    @classmethod
    def from_api_response(cls, info):
        def from_date(s):
            if s:
                return datetime.datetime.strptime(s, "%Y-%m-%dT%H:%M:%SZ").replace(
                    tzinfo=datetime.timezone.utc
                )
            return None

        return cls(
            request_submitted=from_date(info.get("absenteeRequestSubmittedDate")),
            request_approved=from_date(info.get("absenteeRequestApprovedDate")),
            created=from_date(info.get("absenteeBallotCreatedDate")),
            sent=from_date(info.get("absenteeBallotSentDate")),
            expected_delivery=from_date(
                info.get("absenteeBallotAnticipatedDeliveryDate")
            ),
            received=from_date(info.get("completedAbsenteeBallotReceivedDate")),
            returned=from_date(info.get("dateBallotReturned")),
            ballot_rejected=info.get("isDateBallotReturnedRejected"),
            request_denied=info.get("isAbsenteeRequestDenied"),
            ballot_received_issue=info.get("hasAbsenteeBallotReceivedIssue"),
            foreign_address=info.get("isForeignAddress"),
        )


@dataclass
class WIElection:
    election_id: str
    start: datetime.datetime
    end: datetime.datetime
    description: str

    polling_place_ward: str

    polling_place_id: str
    polling_place_description: str
    polling_place_address: str
    polling_place_city: str
    polling_place_state: str
    polling_place_zipcode: str
    polling_place_lat: float
    polling_place_lng: float


@dataclass
class WIVoterRegistration:
    full_name: str
    address: str
    city: str
    state: str
    zipcode: str

    date_of_birth: datetime.date

    status: str
    status_reason: str

    registration_date: datetime.date
    registration_source: str

    voter_reg_number: str  # NOTE: should this be an integer?

    voter_id: str
    district_combo_id: str
    jurisdiction_id: str

    @property
    def active(self):
        return self.status == "Active"

    @property
    def last_name(self):
        return self.full_name.split(",")[0]

    @property
    def first_and_middle_name(self):
        self.full_name.split(",")[1]

    @classmethod
    def from_api_response(cls, info):
        dob = info.get("dateOfBirth")
        if dob and not dob.startswith("01/01/1900"):
            dob = datetime.datetime.strptime(dob[0:10], "%Y-%m-%d").date()
        else:
            dob = None
        r = cls(
            full_name=info.get("voterName"),
            date_of_birth=dob,
            address=info.get("address"),
            city=info.get("city"),
            state=info.get("state"),
            zipcode=info.get("postalCode"),
            status=info.get("voterStatusName"),
            status_reason=info.get("statusReasonName"),
            registration_date=datetime.datetime.strptime(
                info.get("registrationDate"), "%m/%d/%Y"
            ).date(),
            registration_source=info.get("registrationSource"),
            voter_reg_number=info.get("voterRegNumber"),
            voter_id=info.get("voterID"),
            district_combo_id=info.get("districtComboID"),
            jurisdiction_id=info.get("jurisdictionID"),
        )
        return r


def lookup_voter(first_name, last_name, date_of_birth, **kwargs):
    response = requests.post(
        SEARCH_ENDPOINT,
        headers={"content-type": "application/x-www-form-urlencoded",},
        data={
            "firstName": first_name,
            "lastName": last_name,
            "birthDate": date_of_birth.strftime("%m/%d/%Y"),
        },
        **kwargs
    )
    if not response.json().get("Success"):
        return None
    r = []
    for info in response.json()["Data"]["voters"]["$values"]:
        r.append(WIVoterRegistration.from_api_response(info))
    return r


def lookup_polling_place(district_combo_id, **kwargs):
    response = requests.get(
        POLLING_PLACE_ENDPOINT.format(district_combo_id=district_combo_id),
        headers={"content-type": "application/x-www-form-urlencoded",},
        **kwargs
    )
    if not response.json().get("Success"):
        return None
    info = response.json()["Data"]

    # note the "date" is midnight local time in UTC.  we'll adjust the hour/minute and create
    # datetimes with no timezone.
    date = datetime.datetime.strptime(
        info.get("electionDate"), "%Y-%m-%dT%H:%M:%SZ"
    ).replace(tzinfo=datetime.timezone.utc)
    print(date)
    start_time = datetime.datetime.strptime(info.get("startTime"), "%I.%M %p")
    end_time = datetime.datetime.strptime(info.get("endTime"), "%I.%M %p")
    start = date + datetime.timedelta(hours=start_time.hour, minutes=start_time.hour)
    end = date + datetime.timedelta(hours=end_time.hour, minutes=end_time.minute)
    return WIElection(
        election_id=info.get("electionID"),
        start=start,
        end=end,
        description=info.get("electionDescription"),
        polling_place_ward=info.get("wardName"),
        polling_place_id=info.get("pplid"),
        polling_place_description=info.get("pollingLocationName"),
        polling_place_address=info.get("ppL_Address"),
        polling_place_city=info.get("ppL_City"),
        polling_place_state="WI",
        polling_place_zipcode=info.get("ppL_PostalCode"),
        polling_place_lat=float(info.get("latitude")),
        polling_place_lng=float(info.get("longitude")),
    )


def lookup_ballot_status(voter_id, election_id, **kwargs):
    response = requests.get(
        BALLOT_ENDPOINT.format(voter_id=voter_id, election_id=election_id),
        headers={"content-type": "application/x-www-form-urlencoded",},
        **kwargs
    )
    if not response.json().get("Success"):
        return None
    return WIAbsenteeBallotStatus.from_api_response(response.json()["Data"])



"""Version details for ovrlib"""

__title__ = "ovrlib"
__description__ = "Library to interface with per-state online voter registration APIs"
__url__ = "https://github.com/vote/state-registration-apis"
__version__ = "0.0.5"
__author__ = "Sage Weil"
__author_email__ = "sage@newdream.net"
__license__ = "Apache 2.0"
__maintainer__ = "Sage Weil"
__maintainer_email__ = "sage@newdream.net"
__keywords__ = "python ovr vote"



# type: ignore

import os
import textwrap

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

about = {}
with open(os.path.join(here, "ovrlib", "__version__.py"), "r") as f:
    exec(f.read(), about)


setup(
    name=about["__title__"],
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    maintainer=about["__maintainer__"],
    maintainer_email=about["__maintainer_email__"],
    packages=["ovrlib",],
    url=about["__url__"],
    license=about["__license__"],
    description=about["__description__"],
    long_description=textwrap.dedent(open("README.md", "r").read()),
    long_description_content_type="text/markdown",
    install_requires=["requests>=2.22.0", "requests[socks]", "lxml", "dataclasses"],
    tests_require=["pytest", "responses",],
    test_suite="nose.collector",
    keywords=about["__keywords__"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)
